/*========================================================================
     "gpop1scf"  Gaussian output post processor 1
        - read Gaussian SCF-freq job output and create a .gpo file

             Copyright (c) 2002-2018 by A. Miyoshi, Univ. Tokyo
                            GauProc1 created: Oct.  12, 2002
                                      edited: Apr.  26, 2003
                            gpop1scf created: Sept.  1, 2004
                                      edited: Jan.  13, 2005
                                      edited: Aug.   2, 2005
                                 - revised to support #T output
                                      edited: Aug.  22, 2006
                                      edited: Nov.  19, 2007
                                 - changed .mod output file name
                                      edited: Jan.   2, 2009
           - changed .mod defaults: useSCFenergy+ZPE for SCF only job,
                              energyTST CBS-QB3_E(0K) for CBS-QB3 job
           - add .xyz output as an option
                                      edited: Jul.  25, 2009
           - changed template for setIntRotor
                                      edited: Aug.  10, 2009
           - changed comment header in .xyz output from "!" to "#"
                                      edited: Sept.  2, 2009
           - corrected a bug reading CBS-QB3_E(0K)
                                      edited: Nov.  11, 2009
           - changed to read-out the last Freq keysword
                                      edited: Dec.  21, 2009
           - added workaround for NoSymm calculation
                                      edited: Feb.   8, 2010
           - chaged to be compatible with G09 output
                                      edited: Apr.   9, 2010
           - added text summary output
                                      edited: Dec.  17, 2014
           - slightly changed the text summary output
                                      edited: Sept. 10, 2015
           - modified job-end search for Gaussian 16
                                      edited: June   6, 2018
--------------------------------------------------------------------------
 usage: gpop1scf basename[.ext] [-csv/-all]
--------------------------------------------------------------------------
 - Compatible with output of Gaussian 94, 98, and 03.
 - If no extension, default extension '.log' will be added.
 - By default, files basename.gpo and basename.mod are created.
    (If basename.mod is already present, basename_.mod is created.)
    With '-csv' option, 'basename.csv' is also created.
    With '-xyz' option, 'basename.xyz' is also created.
    With '-txt' option, 'basename.txt' is also created.
    With '-all' option, .gpo, .csv, .txt, .frq, .geo, .tst, and .zpe files
     are created.
--------------------------------------------------------------------------
 [examples]
 % gpop1scf methane
  - methane.log is read, methane.gpo is created.
 % gpop1scf ethane.out -csv
  - ethane.out is read, ethane.gpo and ethane.scv are created.
========================================================================*/

#include "gpop1scf.h"

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int ifrq, irot, flgocsv = false, flgotxt = false, flgoall = false, ifj;
  int flgoxyz = false, badopt = false, isCBSQB3job = false, nFreqJobs = 0;
  string gaufln, opt, basfln, extfln, route, comm, gauv, esta, scfe, chml;
  string e0K_cbsqb3;
  ifstream inf;
  vector<string> geom, freq;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;

  showRevision();
  if (argc > 1) { gaufln = argv[1]; }
  if (argc > 2) {
    opt = argv[2];
    if (opt == "-csv") { flgocsv = true; }
    else if (opt == "-xyz") { flgoxyz = true; }
    else if (opt == "-txt") { flgotxt = true; }
    else if (opt == "-all") { flgoall = true; }
    else { badopt = true; }
  }
  if ((argc < 2) || (argc > 3) || ((argc == 3) && badopt))
    { cout << "gpop1scf basename[.ext] [-csv/-xyz/-txt/-all]\n"; return 1; }
   basfln = getBasePart(gaufln); extfln = getExtPart(gaufln);
   if (extfln == "") { gaufln = basfln + ".log"; }

  // --- read-in Gaussian output to count "Freq" in root section
  inf.open(gaufln.c_str());
  if (!inf) { cout << "failed to open [" << gaufln << "].\n"; return 1; }
  if (findJobHead(inf, gauv) == ERRE)
    { cout << "failed to read [" << gaufln << "].\n"; return 1; }
  while (findFreqJob(inf, route, comm) == NORM) { nFreqJobs++; }
  inf.close();
  if (nFreqJobs < 1) { cout << "no frequency job was found.\n"; return 1; }
  inf.clear();

  // --- read in gaussian output
  inf.open(gaufln.c_str());
  if (!inf) { cout << "failed to open [" << gaufln << "].\n"; return 1; }
  if (findJobHead(inf, gauv) == ERRE)
    { cout << "failed to read [" << gaufln << "].\n"; return 1; }
  for (ifj = 0; ifj < nFreqJobs; ifj++) {
    if (findFreqJob(inf, route, comm) == ERRE)
      { cout << "failed to read [" << gaufln << "].\n"; return 1; }
  }
  if (getGeoFrq(inf, geom, freq, esta, scfe, chml) == ERRE)
    { cout << "no frequency output was found.\n"; return 1; }
  isCBSQB3job = getCBSQB3summary(inf, e0K_cbsqb3);
  inf.close();

  // --- process job-infos, molec-grom and freq-table
  jbinf.setFrom(gauv, route, comm, scfe);
  if (molc.setFrom(geom, esta, chml) == ERRE)
    { cout << "geom read error.\n"; return 1; }
  if (frtab.setFrom(freq, jbinf, molc) == ERRE)
    { cout << "freq read error.\n"; return 1; }

  // --- output .geo, .frq, .csv, .tst, .zpe, .gpo, and _.mod
  if (flgoall) {
    if (outFileGeo(basfln, geom) == ERRE) { return 1; }
    if (outFileFrq(basfln, freq, jbinf, molc) == ERRE) { return 1; }
    if (outFileTst(basfln, jbinf, molc, frtab) == ERRE)
      { return 1; }
    if (outFileZpe(basfln, jbinf, frtab) == ERRE) { return 1; }
  }
  if (flgoall || flgocsv) {
    if (outFileCsv(basfln, jbinf, molc, frtab) == ERRE) { return 1; }
  }
  if (flgoall || flgoxyz) {
    if (outFileXyz(basfln, jbinf, molc) == ERRE) { return 1; }
  }
  if (flgoall || flgotxt) {
    if (outFileTxt(basfln, jbinf, molc, frtab, isCBSQB3job,
        e0K_cbsqb3) == ERRE)
      { return 1; }
  }
  if (outFileGpo(basfln, jbinf, molc, frtab, gaufln) == ERRE)
    { return 1; }
  if (outFileModTemplate(basfln, jbinf, molc, frtab, isCBSQB3job,
      e0K_cbsqb3) == ERRE)
    { return 1; }

  return 0;
}

/*===== find gaussian job head =========================================*/

int findJobHead(istream &inf, string &gauv) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 9) == " ********") {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.substr(0, 9) == " Gaussian")
        { gauv = slin; found = true; }
    } else if (slin.substr(0, 19) == " Link1:  Proceeding") {
      found = true;
    }
  } while (!found);
  return NORM;
}

/*===== find frequency job =============================================*/

int findFreqJob(istream &inf, string &route, string &comm) {
  int found = false;
  do {
    if ((route = getRouteSect(inf)) == "") { return ERRE; }
    if (route.find("FREQ") != string::npos) { found = true; }
    else { if (findJobEnd(inf) == ERRE) { return ERRE; } }
  } while (!found);
  comm = getCommLine(inf);
  return NORM;
}

/*===== get geometry and frequencies parts =============================*/

int getGeoFrq(istream &inf, vector<string> &gm, vector<string> &fr,
 string &estate, string &scfene, string &chgmul) {
  int freqfound = false, geomend, freqend, sepc, foundstdg;
  string slin;
  vector<string> gminp;
  vector<string>::iterator vsp;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.find("Input orientation:") != string::npos) {
      // store input-orientation for the case no standard-orientation is found
      gminp.clear(); gminp.push_back(slin); sepc = 0;
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.substr(1,10) == "----------") { sepc++; }
        gminp.push_back(slin);
      } while (sepc < 3);
    }
    if (slin.substr(0, 16) == " Framework group") {
      gm.clear(); gm.push_back(slin); geomend = false;
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.substr(0, 21) == " Rotational constants") {
          geomend = true; foundstdg = false;
          // check whether standard orientation is present
          for (vsp = gm.begin(); vsp != gm.end(); vsp++) {
            if (vsp->substr(1,10) == "----------")
              { foundstdg = true; break; }
          }
          // append input-orientation if standard-orientation not found
          if (!foundstdg) {
            for (vsp = gminp.begin(); vsp != gminp.end(); vsp++)
              { gm.push_back(*vsp); }
          }
        }
        gm.push_back(slin);
      } while (!geomend);
    }
    if (slin.substr(0, 21) == " Harmonic frequencies") {
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
      } while (slin.rfind("normal coordinates:") == string::npos);
      fr.clear(); freqend = false;
      do {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.length() < 22) { freqend = true; }
        else { fr.push_back(slin); }
      } while (!freqend);
      freqfound = true;
    }
    if (slin.find("Multiplicity") != string::npos) { chgmul = slin; }
    if (slin.find("electronic state") != string::npos) { estate = slin; }
    if (slin.find("SCF Done:") != string::npos) { scfene = slin; }
  } while (!freqfound);
  return NORM;
}

/*===== get CBS-QB3 final summary (if exists) ==========================*/

int getCBSQB3summary(istream &inf, string &e0K_cbsqb3) {
  int cbsfound = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return cbsfound; }
    if (slin.substr(0, 15) == " CBS-QB3 (0 K)=") {
      if (slin.substr(39, 7) == "CBS-QB3") {
        // - G09
        e0K_cbsqb3 = trim(slin.substr(20, 18));
      } else {
        e0K_cbsqb3 = trim(slin.substr(20, 16));
      }
      cbsfound = true;
    }
  } while (!cbsfound);
  return cbsfound;
}

/*----- get route section ----------------------------------------------*/

string getRouteSect(istream &inf) {
  int endroute = false;
  string slin, route;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 6) == " -----") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      if (slin.substr(0, 2) == " #") {
        route = slin;
        do {
          if (getLine(inf, slin) == ERRE) { return ""; }
          if (slin.substr(0, 6) == " -----") { endroute = true; }
          else { route += trim(slin); }
        } while (!endroute);
      }
    }
  } while (!endroute);
  strAllToUpper(route);
  return route;
}

/*----- get comment line -----------------------------------------------*/

string getCommLine(istream &inf) {
  int endcomm = false;
  string slin, comm;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 3) == " --") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      comm = slin;
      do {
        if (getLine(inf, slin) == ERRE) { return ""; }
        if (slin.substr(0, 3) == " --") { endcomm = true; }
        else { comm += slin; }
      } while (!endcomm);
    }
  } while (!endcomm);
  if (comm.length() > 0) {
    if (comm[0] == ' ') { comm = comm.substr(1); }
  }
  return comm;
}

/*----- find gaussian job end ------------------------------------------*/

int findJobEnd(istream &inf) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 14) == " Job cpu time:") {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.find("termination") != string::npos) {
        found = true;
      } else {
        if (getLine(inf, slin) == ERRE) { return ERRE; }
        if (slin.find("termination") != string::npos) { found = true; }
      }
    }
  } while (!found);
  return NORM;
}

/*===== output .geo file ===============================================*/

int outFileGeo(string &basfln, vector<string> &geom) {
  string goutfln;
  ofstream gof;
  vector<string>::iterator ps;

  goutfln = basfln + ".geo";
  gof.open(goutfln.c_str());
  if (!gof)
    { cout << "failed to open [" << goutfln << "].\n"; return ERRE; }
  for (ps = geom.begin(); ps != geom.end(); ps++) { gof << *ps << endl; }
  gof.close();
  return NORM;
}

/*===== output .frq file ===============================================*/

int outFileFrq(string &basfln, vector<string> &freq, jobInfos &ji,
 molecGeom &mg) {
  string foutfln;
  ofstream fof;
  vector<string>::iterator ps;

  foutfln = basfln + ".frq";
  fof.open(foutfln.c_str());
  if (!fof)
    { cout << "failed to open [" << foutfln << "].\n"; return ERRE; }
  fof << ji.comment << endl;
  fof << mg.nAtms << endl;
  fof << mg.nVibm << endl;
  for (ps = freq.begin(); ps != freq.end(); ps++) { fof << *ps << endl; }
  fof.close();
  return NORM;
}

/*===== output .xyz file ===============================================*/

int outFileXyz(string &basfln, jobInfos &ji, molecGeom &mg) {
  int prec, iatm, ixyz;
  string xyzfln;
  ofstream xof;

  xyzfln = basfln + ".xyz";
  xof.open(xyzfln.c_str());
  if (!xof)
    { cout << "failed to open [" << xyzfln << "].\n"; return ERRE; }
  prec = xof.precision(); xof.precision(6); xof.setf(ios::fixed);
  xof << mg.nAtms << endl;
  xof << "# " << ji.comment << endl;
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    xof << elm.getAtmcSym(mg.geom[iatm].atomicNumber);
    for (ixyz = 0; ixyz < 3; ixyz++)
      { xof << ' ' << mg.geom[iatm][ixyz]; }
    xof << endl;
  }
  xof.unsetf(ios::fixed); xof.precision(prec);
  xof.close();
  return NORM;
}

/*===== output .csv file ===============================================*/

int outFileCsv(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int ifrq, prec, irot, iatm, ixyz;
  string csvfln;
  ofstream cof;
  map<int, int>::iterator p;

  csvfln = basfln + ".csv";
  cof.open(csvfln.c_str());
  if (!cof)
    { cout << "failed to open [" << csvfln << "].\n"; return ERRE; }

  cof << "[Job informations]\n";
  cof << "version,Gaussian " << ji.version << endl;
  cof << "revision," << ji.revision << endl;
  cof << "comment," << ji.comment << endl;
  cof << "methodIdx," << ji.methodIdx << endl;
  prec = cof.precision(); cof.precision(12);
  cof << "scaleFactVib," << ji.scaleFactVib << endl;
  cof << "scaleFactZPE," << ji.scaleFactZPE << endl;
  cof << "scaleFactGT," << ji.scaleFactGT << endl;
  cof.precision(prec);
  cof << "SCFenergy," << ji.SCFenergy << endl;
  cof << endl;

  cof << "[Geometry and related properties]\n";
  cof << "nAtms," << mg.nAtms << endl;
  cof << "gElec," << mg.gElec << endl;
  cof << "nVibm," << mg.nVibm << endl;
  cof << "isLinear,";
  if (mg.isLinear) { cof << "true\n"; }
  else { cof << "false\n"; }
  cof << "pntGrp," << mg.pntGrp << endl;
  cof << "elState," << mg.elState << endl;
  cof << "rotSymNbr," << mg.rotSymNbr << endl;
  cof << "numIsomers," << mg.numIsomers << endl;
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(6);
  cof << "molW," << mg.molW << endl;
  cof.precision(7);
  cof << "rotcGHz";
  for (irot = 0; irot < 3; irot++) { cof << ',' << mg.rotcGHz[irot]; }
  cof << endl;
  cof << "chemform";
  for (p = mg.chemform.begin(); p != mg.chemform.end(); p++)
    { cof << ',' << elm.getAtmcSym(p->first) << ',' << p->second; }
  cof << endl;
  cof.precision(6);
  cof << "centNbr,atmcNbr,atmcTyp,X,Y,Z\n";
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    cof << iatm + 1 << ',' << mg.geom[iatm].atomicNumber << ','
        << mg.geom[iatm].atomicType;
    for (ixyz = 0; ixyz < 3; ixyz++)
      { cof << ',' << mg.geom[iatm][ixyz]; }
    cof << endl;
  }
  cof.unsetf(ios::fixed); cof.precision(prec);
  cof << endl;

  cof << "[Vibrational frequencies and related properties]\n";
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(4);
  cof << "zpe_Unscaled," << ft.zpe_Unscaled << endl;
  cof << "zpe," << ft.zpe << endl;
  cof << "zpe_GT," << ft.zpe_GT << endl;
  cof << "vibNbr,sym,freq,rMass,fConst,IRint,ramAct,depol\n";
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << "," << ft[ifrq].sym << ",";
    cof.setf(ios::fixed); cof.precision(4);
    cof << ft[ifrq].freq << "," << ft[ifrq].rmass << ","
      << ft[ifrq].fconst << "," << ft[ifrq].irint << ","
      << ft[ifrq].ramact << "," << ft[ifrq].depol << endl;
  }
  cof << "vibNbr,centNbr,dX,dY,dZ\n";
  cof.precision(2);
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << endl;
    for (iatm = 0; iatm < mg.nAtms; iatm++) {
      cof << ',' << iatm + 1;
      for (ixyz = 0; ixyz < 3; ixyz++)
        { cof << ',' << ft[ifrq].nrm[iatm][ixyz]; }
      cof << endl;
    }
  }
  cof.unsetf(ios::fixed); cof.precision(prec);

  cof.close();
  return NORM;
}

/*===== output .txt file ===============================================*/

int outFileTxt(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, int isCBSQB3job, string &e0K_cbsqb3) {
  int ifrq, prec, irot, iatm, ixyz;
  double eSCFzpe = atof(ji.SCFenergy.c_str()) + ft.zpe / WAVNHRT;
  string txtfln;
  ofstream tof;

  txtfln = basfln + ".txt";
  tof.open(txtfln.c_str());
  if (!tof)
    { cout << "failed to open [" << txtfln << "].\n"; return ERRE; }

  tof << "name, " << ji.comment << endl;
  tof << "chemical formula, " << getChemForm(mg.chemform) << endl;
  tof << "degeneracy of electronic state, " << mg.gElec << endl;
  tof << "point group, " << mg.pntGrp << endl;
  tof << "electronic state, " << mg.elState << endl;
  tof << "rotational symmetry number, " << mg.rotSymNbr << endl;
  tof << "number of optical isomers, " << mg.numIsomers << endl;
  prec = tof.precision(); tof.setf(ios::fixed); tof.precision(6);
  tof << "molecular weight, " << mg.molW << endl;
  tof << "CBS-QB3 E(0K) [hartree], ";
  if (isCBSQB3job) { tof << e0K_cbsqb3 << endl; }
  else { tof << "-\n"; }
  tof << "SCFenergy, " << ji.SCFenergy << endl;
  tof.precision(10);
  tof << "SCFenergy+ZPE, " << eSCFzpe << endl;
  tof.precision(7);
  tof << "rotational constants [GHz]";
  for (irot = 0; irot < 3; irot++) { tof << ", " << mg.rotcGHz[irot]; }
  tof << endl;

  tof.precision(6);
  tof << "atomic symbol, x, y, z [Angstrom]\n";
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    tof << elm.getAtmcSym(mg.geom[iatm].atomicNumber);
    for (ixyz = 0; ixyz < 3; ixyz++)
      { tof << ", " << mg.geom[iatm][ixyz]; }
    tof << endl;
  }
  tof.unsetf(ios::fixed); tof.precision(prec);

  prec = tof.precision(); tof.setf(ios::fixed); tof.precision(4);
  tof << "symmetry species, frequency [cm-1]\n";
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    tof << ft[ifrq].sym << ", ";
    tof.setf(ios::fixed); tof.precision(4);
    tof << ft[ifrq].freq << endl;
  }

  tof.close();
  return NORM;
}

/*===== output .zpe file ===============================================*/

int outFileZpe(string &basfln, jobInfos &ji, freqTable &ft) {
  string zoutfln;
  ofstream zof;

  zoutfln = basfln + ".zpe";
  zof.open(zoutfln.c_str());
  if (!zof)
    { cout << "failed to open [" << zoutfln << "].\n"; return ERRE; }
  zof << "method detected as " << ji.method << "/" << ji.basis << ".\n";
  zof << "scaling factors for Vib.Freq. / ZPE / Gau.Theory :\n";
  zof << " " << ji.scaleFactVib << " / " << ji.scaleFactZPE << " / "
      << ji.scaleFactGT << endl;
  zof.setf(ios::fixed); zof.precision(4);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /cm-1\n";
  zof << " " << ft.zpe_Unscaled << "  " << ft.zpe << "  " << ft.zpe_GT
      << endl;
  zof.precision(10);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /hartree\n";
  zof << " " << ft.zpe_Unscaled / WAVNHRT << "  " << ft.zpe / WAVNHRT
      << "  " << ft.zpe_GT / WAVNHRT << endl;
  zof.close();
  return NORM;
}

/*===== output .mod template file ======================================*/

int outFileModTemplate(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, int isCBSQB3job, string &e0K_cbsqb3) {
  int isTS = false;
  string modfln;
  ifstream inf;
  ofstream mof;

  modfln = basfln + ".mod";
  inf.open(modfln.c_str());
  if (inf) { modfln = basfln + "_.mod"; }
  inf.close();
  mof.open(modfln.c_str());
  if (!mof)
    { cout << "failed to open [" << modfln << "].\n"; return ERRE; }
  mof << "! gElec " << mg.gElec << endl;
  mof << "! rotSymNbr " << mg.rotSymNbr << endl;
  mof << "! numIsomers " << mg.numIsomers << endl;
  if (ft[0].freq < 0.) { isTS = true; }
  mof << "! isTS " << strGpoBlnv(isTS) << endl;
  if (isCBSQB3job) {
    mof << "! ## CBS-QB3 E(0K) is used as default. Change if necessary.\n";
    mof << "energyTST " << e0K_cbsqb3 << endl;
    mof << "! useSCFenergy+ZPE\n";
  } else {
    mof << "! energyTST #\n";
    mof << "! ## remove or comment out the next line when you set enegyTST.\n";
    mof << "useSCFenergy+ZPE\n";
  }
  mof << "! setIntRotor idVib nSym moi1 moi2 V0 comment\n";
  mof.close();
  return NORM;
}

