/* =======================================================================
     "gpop2mlp"  Gaussian output post processor 2
        - read Molpro frequency job output and create a .gpo file

             Copyright (c) 2002-2013 by A. Miyoshi, Hiroshima Univ.
                                     created: Jan.  27, 2013
                                 last edited: Sep.  17, 2020
--------------------------------------------------------------------------
 usage: gpop2mlp basename[.ext] [-csv/-xyz/-all]
--------------------------------------------------------------------------
 - Supports output of Molpro 2002.6, 2008.1 and 2010.1.
 - If no extension is specified, default extension '.out' will be assumed.
 - Files named basename.gpo and basename.mod are created.
   If basename.mod already exists, basename_.mod is created.
======================================================================= */

#include "gpop2mlp.h"

/* ===== main function ================================================ */

int main(int argc, char *argv[]) {
  int flgocsv = false, flgoxyz = false, flgoall = false, badopt = false;
  string mlpfln, opt, basfln, extfln;
  string mlpv, ptch, comm, bsis, meth, mult, scfe;
  ifstream inf;
  vector<string> geom, freq;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;

  showRevision();
  if (argc > 1) { mlpfln = argv[1]; }
  if (argc > 2) {
    opt = argv[2];
    if (opt == "-csv") { flgocsv = true; }
    else if (opt == "-xyz") { flgoxyz = true; }
    else if (opt == "-all") { flgoall = true; }
    else { badopt = true; }
  }
  if ((argc < 2) || (argc > 3) || ((argc == 3) && badopt))
    { cout << "gpop2mlp basename[.ext] [-csv/-xyz/-all]\n"; return 1; }
   basfln = getBasePart(mlpfln); extfln = getExtPart(mlpfln);
   if (extfln == "") { mlpfln = basfln + ".out"; }

  // --- read in Molpro output
  inf.open(mlpfln.c_str());
  if (!inf) { cout << "failed to open [" << mlpfln << "].\n"; return 1; }
  if (findJobHead(inf, mlpv, ptch, comm) == ERRE)
    { cout << "failed to read header [" << mlpfln << "].\n"; return 1; }
  if (findFreqJob(inf, bsis, meth, mult) == ERRE)
    { cout << "failed to read freq job [" << mlpfln << "].\n"; return 1; }
  if (getGeoFrq(inf, geom, freq) == ERRE)
    { cout << "no frequency output was found.\n"; return 1; }
  if (findJobEnd(inf, scfe) == ERRE)
    { cout << "failed to read terminator [" << mlpfln << "].\n"; return 1; }
  inf.close();

  // --- process jobInfos, molecGeom and freqTable
  jbinf.setFromMlp(mlpv, ptch, bsis, meth, comm, scfe);
  if (molc.setFromMlp(geom, mult) == ERRE)
    { cout << "geom read error.\n"; return 1; }
  if (frtab.setFromMlp(freq, jbinf, molc) == ERRE)
    { cout << "freq read error.\n"; return 1; }

  // --- output .geo, .frq, .csv, .tst, .zpe, .gpo, and _.mod
  if (flgoall) {
    if (outFileGeo(basfln, geom) == ERRE) { return 1; }
    if (outFileFrq(basfln, freq, jbinf, molc) == ERRE) { return 1; }
    if (outFileTst(basfln, jbinf, molc, frtab) == ERRE)
      { return 1; }
    if (outFileZpe(basfln, jbinf, frtab) == ERRE) { return 1; }
  }
  if (flgoall || flgocsv) {
    if (outFileCsv(basfln, jbinf, molc, frtab) == ERRE) { return 1; }
  }
  if (flgoall || flgoxyz) {
    if (outFileXyz(basfln, jbinf, molc) == ERRE) { return 1; }
  }
  if (outFileGpo(basfln, jbinf, molc, frtab, mlpfln) == ERRE)
    { return 1; }
  if (outFileModTemplate(basfln, jbinf, molc, frtab) == ERRE)
    { return 1; }

  return 0;
}

/* ===== find Molpro job header ======================================= */

int findJobHead(istream &inf, string &mlpv, string &ptch, string &comm) {
  string slin;
  vector<string> tkns;
  // - header
  if (skipLinesToKey(inf, slin, "PROGRAM SYSTEM MOLPRO") == ERRE)
    { return ERRE; }
  // - version
  // changed for version 2015: Feb. 27, 2017
  if (skipGetLine(inf, 1, slin) == ERRE) { return ERRE; }
  tkns = splitTokens(slin, " ");
  if (tkns.size() > 1) {
    if (tkns[0] == "Version") { mlpv = tkns[1]; } else { return ERRE; }
  } else {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    tkns = splitTokens(slin, " ");
    if (tkns.size() > 1) {
      if (tkns[0] == "Version") { mlpv = tkns[1]; } else { return ERRE; }
    } else { return ERRE; }
  }
  // - label
  if (skipGetLine(inf, 3, slin) == ERRE) { return ERRE; }
  if (slin.find("LABEL *") != string::npos) { comm = trim(slin.substr(9)); }
  else { return ERRE; }
  // - patch level
  if (skipGetLine(inf, 3, slin) == ERRE) { return ERRE; }
  if (slin.find("Patch level:") != string::npos) { ptch = trim(slin.substr(14)); }
  else { ptch = "n.a."; }
  return NORM;
}

/* ===== find frequency job =========================================== */

int findFreqJob(istream &inf, string &bsis, string &meth, string &mult) {
  int found = false;
  string slin;
  bsis = "n.a.";
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.find("SETTING BASIS") != string::npos) {
      bsis = trim(slin.substr(26));
    } else if (slin.find("Reference symmetry:") != string::npos) {
      mult = trim(slin.substr(42));
    } else {
      if (slin.find("FREQUENCIES * CALCULATION OF NORMAL MODES FOR")
           != string::npos) {
        meth = trim(slin.substr(47));
        found = true;
      }
    }
  } while (!found);
  return NORM;
}

/* ===== get geometry and frequencies outputs ========================= */

int getGeoFrq(istream &inf, vector<string> &gm, vector<string> &fr) {
  int nblin, freqEnd = false;
  string slin;
  // - geometry
  if (skipLinesToKey(inf, slin, "Atomic Coordinates") == ERRE)
    { return ERRE; }
  if (skipLinesToKey(inf, slin, "Charge") == ERRE) { return ERRE; }
  gm.clear(); gm.push_back(slin); nblin = 0;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (trim(slin) == "") { nblin++; }
    else { gm.push_back(slin); }
  } while (nblin < 2);
   // - frequency
  fr.clear();
  if (skipLinesToKey(inf, slin, "Normal Modes") == ERRE) { return ERRE; }
  fr.push_back(slin); nblin = 0;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (trim(slin) == "") { nblin++; } else { nblin = 0; }
    fr.push_back(slin);
    if (nblin > 1) {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.find("Zero point energy:") != string::npos)
        { freqEnd = true; }
      else if (slin.find("Frequencies dumped") != string::npos)
        { freqEnd = true; }
      else { nblin = 0; fr.push_back(slin); }
    }
  } while (!freqEnd);
  return NORM;
}

/* ===== find Molpro job end ========================================== */

int findJobEnd(istream &inf, string &scfe) {
  int found = false;
  string slin = "", slin1 = "", slin2 = "";
  do {
    slin2 = slin1; slin1 = slin;
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.find("Variable memory released") != string::npos)
      { found = true; }
    else if (slin.find("Molpro calculation terminated") != string::npos)
      { found = true; }
  } while (!found);
  scfe = trim(slin2.substr(1, 15));
  return NORM;
}

/*===== output .geo file ===============================================*/

int outFileGeo(string &basfln, vector<string> &geom) {
  string goutfln;
  ofstream gof;
  vector<string>::iterator ps;

  goutfln = basfln + ".geo";
  gof.open(goutfln.c_str());
  if (!gof)
    { cout << "failed to open [" << goutfln << "].\n"; return ERRE; }
  for (ps = geom.begin(); ps != geom.end(); ps++) { gof << *ps << endl; }
  gof.close();
  return NORM;
}

/*===== output .frq file ===============================================*/

int outFileFrq(string &basfln, vector<string> &freq, jobInfos &ji,
 molecGeom &mg) {
  string foutfln;
  ofstream fof;
  vector<string>::iterator ps;

  foutfln = basfln + ".frq";
  fof.open(foutfln.c_str());
  if (!fof)
    { cout << "failed to open [" << foutfln << "].\n"; return ERRE; }
  fof << ji.comment << endl;
  fof << mg.nAtms << endl;
  fof << mg.nVibm << endl;
  for (ps = freq.begin(); ps != freq.end(); ps++) { fof << *ps << endl; }
  fof.close();
  return NORM;
}

/*===== output .xyz file ===============================================*/

int outFileXyz(string &basfln, jobInfos &ji, molecGeom &mg) {
  int prec, iatm, ixyz;
  string xyzfln;
  ofstream xof;

  xyzfln = basfln + ".xyz";
  xof.open(xyzfln.c_str());
  if (!xof)
    { cout << "failed to open [" << xyzfln << "].\n"; return ERRE; }
  prec = xof.precision(); xof.precision(6); xof.setf(ios::fixed);
  xof << mg.nAtms << endl;
  xof << "# " << ji.comment << endl;
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    xof << elm.getAtmcSym(mg.geom[iatm].atomicNumber);
    for (ixyz = 0; ixyz < 3; ixyz++)
      { xof << ' ' << mg.geom[iatm][ixyz]; }
    xof << endl;
  }
  xof.unsetf(ios::fixed); xof.precision(prec);
  xof.close();
  return NORM;
}

/*===== output .csv file ===============================================*/

int outFileCsv(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int ifrq, prec, irot, iatm, ixyz;
  string csvfln;
  ofstream cof;
  map<int, int>::iterator p;

  csvfln = basfln + ".csv";
  cof.open(csvfln.c_str());
  if (!cof)
    { cout << "failed to open [" << csvfln << "].\n"; return ERRE; }

  cof << "[Job informations]\n";
  cof << "version," << ji.version << endl;
  cof << "revision," << ji.revision << endl;
  cof << "comment," << ji.comment << endl;
  cof << "methodIdx," << ji.methodIdx << endl;
  prec = cof.precision(); cof.precision(12);
  cof << "scaleFactVib," << ji.scaleFactVib << endl;
  cof << "scaleFactZPE," << ji.scaleFactZPE << endl;
  cof << "scaleFactGT," << ji.scaleFactGT << endl;
  cof.precision(prec);
  cof << "SCFenergy," << ji.SCFenergy << endl;
  cof << endl;

  cof << "[Geometry and related properties]\n";
  cof << "nAtms," << mg.nAtms << endl;
  cof << "gElec," << mg.gElec << endl;
  cof << "nVibm," << mg.nVibm << endl;
  cof << "isLinear,";
  if (mg.isLinear) { cof << "true\n"; }
  else { cof << "false\n"; }
  cof << "pntGrp," << mg.pntGrp << endl;
  cof << "elState," << mg.elState << endl;
  cof << "rotSymNbr," << mg.rotSymNbr << endl;
  cof << "numIsomers," << mg.numIsomers << endl;
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(6);
  cof << "molW," << mg.molW << endl;
  cof.precision(7);
  cof << "rotcGHz";
  for (irot = 0; irot < 3; irot++) { cof << ',' << mg.rotcGHz[irot]; }
  cof << endl;
  cof << "chemform";
  for (p = mg.chemform.begin(); p != mg.chemform.end(); p++)
    { cof << ',' << elm.getAtmcSym(p->first) << ',' << p->second; }
  cof << endl;
  cof.precision(9);
  cof << "centNbr,atmcNbr,atmcTyp,X,Y,Z\n";
  for (iatm = 0; iatm < mg.nAtms; iatm++) {
    cof << iatm + 1 << ',' << mg.geom[iatm].atomicNumber << ','
        << mg.geom[iatm].atomicType;
    for (ixyz = 0; ixyz < 3; ixyz++)
      { cof << ',' << mg.geom[iatm][ixyz]; }
    cof << endl;
  }
  cof.unsetf(ios::fixed); cof.precision(prec);
  cof << endl;

  cof << "[Vibrational frequencies and related properties]\n";
  prec = cof.precision(); cof.setf(ios::fixed); cof.precision(4);
  cof << "zpe_Unscaled," << ft.zpe_Unscaled << endl;
  cof << "zpe," << ft.zpe << endl;
  cof << "zpe_GT," << ft.zpe_GT << endl;
  cof << "vibNbr,sym,freq,rMass,fConst,IRint,ramAct,depol\n";
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << "," << ft[ifrq].sym << ",";
    cof.setf(ios::fixed); cof.precision(4);
    cof << ft[ifrq].freq << "," << ft[ifrq].rmass << ","
      << ft[ifrq].fconst << "," << ft[ifrq].irint << ","
      << ft[ifrq].ramact << "," << ft[ifrq].depol << endl;
  }
  cof << "vibNbr,centNbr,dX,dY,dZ\n";
  cof.precision(5);
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    cof << ifrq + 1 << endl;
    for (iatm = 0; iatm < mg.nAtms; iatm++) {
      cof << ',' << iatm + 1;
      for (ixyz = 0; ixyz < 3; ixyz++)
        { cof << ',' << ft[ifrq].nrm[iatm][ixyz]; }
      cof << endl;
    }
  }
  cof.unsetf(ios::fixed); cof.precision(prec);

  cof.close();
  return NORM;
}

/*===== output .zpe file ===============================================*/

int outFileZpe(string &basfln, jobInfos &ji, freqTable &ft) {
  string zoutfln;
  ofstream zof;

  zoutfln = basfln + ".zpe";
  zof.open(zoutfln.c_str());
  if (!zof)
    { cout << "failed to open [" << zoutfln << "].\n"; return ERRE; }
  zof << "method detected as " << ji.method << "/" << ji.basis << ".\n";
  zof << "scaling factors for Vib.Freq. / ZPE / Gau.Theory :\n";
  zof << " " << ji.scaleFactVib << " / " << ji.scaleFactZPE << " / "
      << ji.scaleFactGT << endl;
  zof.setf(ios::fixed); zof.precision(4);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /cm-1\n";
  zof << " " << ft.zpe_Unscaled << "  " << ft.zpe << "  " << ft.zpe_GT
      << endl;
  zof.precision(10);
  zof << "- ZPE(unscaled)  ZPE  ZPE(for Gaussian Theory) /hartree\n";
  zof << " " << ft.zpe_Unscaled / WAVNHRT << "  " << ft.zpe / WAVNHRT
      << "  " << ft.zpe_GT / WAVNHRT << endl;
  zof.close();
  return NORM;
}

/*===== output .mod template file ======================================*/

int outFileModTemplate(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int isTS = false;
  string modfln;
  ifstream inf;
  ofstream mof;

  modfln = basfln + ".mod";
  inf.open(modfln.c_str());
  if (inf) { modfln = basfln + "_.mod"; }
  inf.close();
  mof.open(modfln.c_str());
  if (!mof)
    { cout << "failed to open [" << modfln << "].\n"; return ERRE; }
  mof << "! gElec " << mg.gElec << endl;
  mof << "! pntGrp " << mg.pntGrp << endl;
  mof << "! elState " << mg.elState << endl;
  mof << "! rotSymNbr " << mg.rotSymNbr << endl;
  mof << "! numIsomers " << mg.numIsomers << endl;
  if (ft[0].freq < 0.) { isTS = true; }
  mof << "! isTS " << strGpoBlnv(isTS) << endl;
  mof << "! energyTST #\n";
  mof << "! ## remove or comment out the next line when you set enegyTST.\n";
  mof << "useSCFenergy+ZPE\n";
  mof << "! setIntRotor idVib nSym moi1 moi2 V0 comment\n";
  mof.close();
  return NORM;
}

