/*========================================================================
     "gpop4thf"  Gaussian output post processor 4
        - calculate thermodynamic functions

             Copyright (c) 2002-2013 by A. Miyoshi, Univ. Tokyo
                                     created: Dec.  28, 2004
                                      edited: Jan.  13, 2005
                                      edited: Nov.  19, 2007
                                      edited: May   30, 2011
                       - added .lst output for THERMFIT
                                      edited: July  15, 2013
                       - separated outFileLst function
                       - added .i97 output for PAC99
                                 last edited: Aug.  12, 2013
                       - corrected a bug in .97 output
--------------------------------------------------------------------------
 usage: gpop4thf basename[.gpo/.tst] [temperat-file]
--------------------------------------------------------------------------
 - The 'file' must be .gpo or .tst file.
 - Keywords allowed in the temperat-file (with any extension) are:
    tempRange Tstart(double) Tend(double) Tstep(double)
    tempList T1(double) T2(double) ...
 - Default temperatures are 300 to 1500 K with 100 K step.
 - If a file named basename.ath exists, this program generates
   a file named basename.lst for THERMFIT
========================================================================*/

#include "gpoplibt.h"

int outFileLst(string &basfln, string &spcnam, molPropTST &mptst,
 molecGeom &molc, double DH298, int numIntRot);
int outFileI97(string &basfln, string &spcnam, molPropTST &mptst,
 molecGeom &molc, double DH298);

/*===== main routine ===================================================*/

int main(int argc, char *argv[]) {
  int useTempFile = false, outLstFile = false, DH298found = false;
  int numIntRot = 0;
  double DH298;
  string flname, basfln, extfln, tprfln, csvfln;
  string auxfln, slin, key, spcnam = "";
  ifstream aif;
  ofstream ouf;
  vector<string> atkn;
  molPropTST mptst;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;
  modInfoTST miTST;
  tempList temperat;
  resMolTherm rmth;

  showRevision();
  if ((argc < 2) || (argc > 3)) {
    cout << "usage: gpop4thf basename[.gpo/.tst] [temperat-file]\n";
    return 1;
  }
  flname = argv[1]; basfln = getBasePart(flname);
  extfln = getExtPart(flname); strAllToLower(extfln);
  if (extfln == "") { extfln = "gpo"; }
  if (argc == 3) { tprfln = argv[2]; useTempFile = true; }
  if (extfln == "tst") {
    if (mptst.readFromTST(basfln) == ERRE)
      { cout << "failed to read [" << flname << "].\n"; return 1; }
  } else if (extfln == "gpo") {
    if (readGpo(basfln, jbinf, molc, frtab, miTST) == ERRE)
      { cout << "failed to read [" << flname << "].\n"; return 1; }
    mptst.setFrom(basfln, jbinf, molc, frtab, miTST);
  } else { cout << "invalid file name.\n"; return 1; }
  if (useTempFile) {
    if (temperat.readFrom(tprfln) == ERRE)
      { cout << "failed reading [" << tprfln << "].\n"; return 1; }
  }

  auxfln = basfln + ".ath";
  aif.open(auxfln.c_str());
  if (aif) {
    outLstFile = true;
    while (getLineGpop(aif, slin) == NORM) {
      atkn = splitTokens(slin, " ,\t");
      if (atkn.size() < 2)
        { cout << "Wrong aux-input.\n"; aif.close(); return 1; }
      key = atkn[0];
      if (key == "speciesName") {
        spcnam = atkn[1];
      } else if (key == "DH298_kJ") {
        DH298 = atof(atkn[1].c_str()); DH298found = true;
      } else if (key == "DH298_RT") {
        DH298 = atof(atkn[1].c_str()) * RGASC * STDTMPR / 1000.;
        DH298found = true;
      } else if (key == "numIntRotors") {
        numIntRot = atoi(atkn[1].c_str());
      } else { cout << "Warning: Invalid key [" << key << "].\n"; }
    }
    aif.close();
    if (spcnam.length() < 1) { spcnam = basfln; }
    if (spcnam.length() > 18) { spcnam = spcnam.substr(0, 18); }
    if (!DH298found) {
      cout << "Warning: No DH298 found in .ath file.\n";
      cout << "         .lst file will not be generated.\n";
      outLstFile = false;
    }
  }

  rmth = mptst.getMolTherm(temperat);
  csvfln = basfln + ".csv";
  ouf.open(csvfln.c_str());
  if (!ouf) { cout << "failed to open [" << csvfln << "].\n"; return 1; }
  rmth.printHeadCSV(ouf); rmth.ref.printCSV(ouf);
  rmth.printHeadCSVs(ouf); rmth.printCSVsubtH(ouf);
  ouf.close();

  if (outLstFile) {
    if (outFileLst(basfln, spcnam, mptst, molc, DH298, numIntRot) == ERRE)
      { return 1; }
    if (outFileI97(basfln, spcnam, mptst, molc, DH298) == ERRE)
      { return 1; }
  }

  return 0;
}

/*===== output .lst file ===============================================*/

int outFileLst(string &basfln, string &spcnam, molPropTST &mptst,
 molecGeom &molc, double DH298, int numIntRot) {
  int i;
  double S298;
  string lstfln = basfln + ".lst", strspc, strelm;
  ofstream lso;
  map<string, int> chemFml;
  map<int, int>::iterator p;
  map<string, int>::iterator q;
  tempList temperLST;
  resMolTherm rmth;

  temperLST.setThermDefault();
  rmth = mptst.getMolTherm(temperLST);
  lso.open(lstfln.c_str());
  if (!lso)
    { cout << "failed to open [" << lstfln << "].\n"; return ERRE; }
  strspc = spcnam + "                 "; strspc = strspc.substr(0, 18);
  S298 = rmth.ref.S - RGASC * log(1.01325);
  chemFml.clear();
  for (p = molc.chemform.begin(); p != molc.chemform.end(); p++)
    { chemFml.insert(make_pair(elm.getAtmcSym(p->first), p->second)); }
  lso << "  UNITS:KJ\n";
  lso << " Thermodata generated by gpop\n";
  lso << " SPECIES       Hf       S    Cp 300     400     500     600     800     1000     1500     DATE        ELEMENTS\n";
  lso << ' ' << strspc.substr(0, 10) << formatF(DH298, 8, 2) << "  "
      << formatF(S298, 7, 2) << "  ";
  for (i = 0; i < rmth.size(); i++)
    { lso << formatF(rmth[i].Cp, 7, 2) << ' '; }
  lso << strspc.substr(10, 8) << " gpop    ";
  q = chemFml.begin();
  for (i = 0; i < 4; i++) {
    if (i >= chemFml.size()) { lso << "    0 "; }
    else {
      strelm = q->first;
      if (strelm.length() == 1) { strelm = strelm + " "; }
      lso << strelm << format(q->second, 3, ' ') << ' ';
      q++;
    }
  }
  lso << "G" << format(numIntRot, 2, ' ') << endl;
  lso.close();
  return NORM;
}

/*===== output .i97 file ===============================================*/

int outFileI97(string &basfln, string &spcnam, molPropTST &mptst,
 molecGeom &molc, double DH298) {
  int i;
  string i97fln = basfln + ".i97", chFormula = "";
  ofstream pao;
  map<string, int> chemFml;
  map<int, int>::iterator p;
  map<string, int>::iterator q;
  tempList temperI97;
  resMolTherm rmth;

  temperI97.setPAC99Default();
  rmth = mptst.getMolTherm(temperI97);
  pao.open(i97fln.c_str());
  if (!pao)
    { cout << "failed to open [" << i97fln << "].\n"; return ERRE; }
  chemFml.clear();
  for (p = molc.chemform.begin(); p != molc.chemform.end(); p++)
    { chemFml.insert(make_pair(elm.getAtmcSym(p->first), p->second)); }
  for (q = chemFml.begin(); q != chemFml.end(); q++)
    { chFormula += q->first + format(q->second); }
  chFormula += "                        ";
  pao << "NAME  " << spcnam << endl;
  pao << chFormula.substr(0, 24) << "HF298 "
      << formatF(DH298 * 1000., 12, 1) << "JOULES\n";
  pao << "DATE  gpop\n";
  pao << "LSTS  OLD\n";
  pao << "OUTP  MFIG              LSQS              ATM\n";
  pao << "METH  READIN            KJOULE            BAR\n";
  for (i = 0; i < rmth.size(); i++) {
    pao << "      T     " << formatF(rmth[i].T, 12, 3)
        << "CP/R  " << formatF(rmth[i].Cp / RGASC, 12, 4)
        << "H-H0RT" << formatF(rmth[i].H / RGASC / rmth[i].T, 12, 7)
        << "S/R   " << formatF(rmth[i].S / RGASC, 12, 4) << endl;
  }
  pao << "FINISH\n";
  pao.close();
  return NORM;
}

