/*========================================================================
     "gpop6irt"  Gaussian output post processor 6
        - calculate reduced moment of inertia of a internal rotor

             Copyright (c) 2002-2013 by A. Miyoshi, Univ. Tokyo
                                     created: Sept.  6, 2004
                                      edited: Jan.  21, 2005
                                      edited: Nov.  19, 2007
--------------------------------------------------------------------------
 usage: gpop6irt basename moi1 moi2
--------------------------------------------------------------------------
 - Read basename.gpo, output is made to console (standard output).
 - moi1 and moi2 are the molecular moiety specification by atom numbers,
    e.g., '3-6-7-8' specifies moiety consist of atom 3, 6, 7, and 8,
    and the pivot atom of internal rotation is atom 3.
 - In many cases, moi2 needs to specifie all the rest, thus the pivot
    atom number with -@, e.g., 2-@ is enough.
--------------------------------------------------------------------------
 [Example]
 % gpop6irt ethane 1-3-4-5 2-@
  - read ethane.gpo, calculate internal CH3 torsion
    (in this case, atom 1 and 2 are carbon, 3 to 8 are hydrogen, and
     hydrogen 3, 4, and 5 are bonded to carbon 1)
========================================================================*/

#include "gpoplibc.h"

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int im, ir;
  string basfln;
  vector<string> moin;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;
  molByMoieties mbm;
  vector<pair<int, double> > modr;

  showRevision();
  if (argc != 4)
    { cout << "usage: gpop6irt basename moi1 moi2\n"; return 1; }
  basfln = argv[1];
  for (im = 0; im < 2; im++) { moin.push_back(argv[im + 2]); }

  // --- read in gpop (.gpo) file
  if (readGpo(basfln, jbinf, molc, frtab) == ERRE) { return 1; }

  // --- get moiety input
  if (mbm.setList(moin, molc) == ERRE)
    { cout << "Error in moiety input.\n"; return 1; }

  // --- calculate moment of intertia
  mbm.setProp(molc);

  // --- output results
  cout << "base file name: " << basfln << endl;
  cout << mbm;

  // --- most resembling vibrational modes
  mbm.setIntRotDevn(molc);
  modr = frtab.getMostAlike(mbm.intRotDevn);
  cout << "[most resembling vibrations]\n";
  for (ir = 0; ir < modr.size(); ir++) {
    cout << " mode-" << format(modr[ir].first + 1, 3, '0') << ": "
     << modr[ir].second << endl;
  }
  return 0;
}

