/*========================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2013 by A. Miyoshi, Univ. Tokyo
                                 last edited: Oct.  26, 2005
========================================================================*/

#include "gpoplibr.h"

/*========================================================================
     Classes for molecular states
========================================================================*/

/*------------------------------------------------------------------------
     Eigen state
------------------------------------------------------------------------*/

eigenState::eigenState() { clear(); }             //---- (constructor) ---

void eigenState::clear()                          //------------ clear ---
 { deg = 1; energy = 0.; }

int eigenState::restoreFrom(istream &is) {        //------ restoreFrom ---
  string slin;
  vector<string> tkns;
  if (getLineGpop(is, slin) == ERRE) { return ERRE; }
  tkns = splitTokens(slin, " ");
  if (tkns.size() < 2) { return ERRE; }
  deg = atoi(tkns[0].c_str());
  energy = atof(tkns[1].c_str());
  return NORM;
}

int eigenState::restoreFromTST(istream &is) {     //--- restoreFromTST ---
  string slin;
  vector<string> tkns;
  if (getLineTrim(is, slin) == ERRE) { return ERRE; }
  tkns = splitTokens(slin, " ");
  if (tkns.size() < 2) { return ERRE; }
  deg = atoi(tkns[0].c_str());
  energy = atof(tkns[1].c_str());
  return NORM;
}

ostream &eigenState::print(ostream &os) {         //------------ print ---
  os << "    " << deg << " " << energy << endl;
  return os;
}

string eigenState::outStrTST() {                  //-------- outStrTST ---
  ostringstream oss;
  oss.setf(ios::fixed); oss.precision(4);
  oss << " " << deg << " " << energy << endl;
  return oss.str();
}

ostream &operator<<(ostream &os, eigenState &es)  //------ operator << ---
  { return es.print(os); }

/*------------------------------------------------------------------------
     List of eigenstates
------------------------------------------------------------------------*/

int listStates::restoreFrom(istream &is, int nes) {  //--- restoreFrom ---
  int ies;
  eigenState es;
  clear();
  for (ies = 0; ies < nes; ies++) {
    if (es.restoreFrom(is) == ERRE) { return ERRE; }
    push_back(es);
  }
  return NORM;
}

                                                  //--- restoreFromTST ---
int listStates::restoreFromTST(istream &is, int nes) {
  int ies;
  eigenState es;
  clear();
  for (ies = 0; ies < nes; ies++) {
    if (es.restoreFromTST(is) == ERRE) { return ERRE; }
    push_back(es);
  }
  return NORM;
}

void listStates::setOneState(int deg) {           //------ setOneState ---
  double energy = 0.;
  eigenState es;
  es.deg = deg; es.energy = energy;
  push_back(es);
}

ostream &listStates::print(ostream &os) {         //------------ print ---
  int ies, nes = size();
  os.setf(ios::fixed); os.precision(4);
  for (ies = 0; ies < nes; ies++) { os << (*this)[ies]; }
  return os;
}

string listStates::outStrTST() {                  //-------- outStrTST ---
  int ies, nes = size();
  string ostr;
  ostr.erase();
  for (ies = 0; ies < nes; ies++) { ostr += (*this)[ies].outStrTST(); }
  return ostr;
}

double listStates::statesQ(double T) {            //---------- statesQ ---
  double qst;
  if (size() == 0) { qst = 1.; }
  else { qst = statesSum(0, T); }
  return qst;
}

double listStates::statesU(double T) {            //---------- statesU ---
  double ust;
  if (size() == 0) { ust = 0.; }
  else { ust = statesSum(1, T) / statesSum(0, T); }
  return ust;
}

double listStates::statesS(double T) {            //---------- statesS ---
  double sst, s0;
  if (size() == 0) { sst = 0.; }
  else { s0 = statesSum(0, T); sst = statesSum(1, T) / s0 + log(s0); }
  return sst;
}

double listStates::statesCv(double T) {           //--------- statesCv ---
  double cvst, s0;
  if (size() == 0) { cvst = 0.; }
  else {
    s0 = statesSum(0, T);
    cvst = statesSum(2, T) / s0 - pow(statesSum(1, T) / s0, 2);
  }
  return cvst;
}

double listStates::statesSum(int n, double T) {   //-------- statesSum ---
  int ies, nes = size(), i;
  double x, term, retv = 0.;
  for (ies = 0; ies < nes; ies++) {
    x = (*this)[ies].energy * KELWAVN / T;
    term = (*this)[ies].deg * exp(-x);
    for (i = 0; i < n; i++) { term *= x; }
    retv += term;
  }
  return retv;
}

ostream &operator<<(ostream &os, listStates &ls)  //------ operator << ---
  { return ls.print(os); }


/*========================================================================
     Classes for molecular-property modification
========================================================================*/

/*------------------------------------------------------------------------
     Decoupled Rotor
------------------------------------------------------------------------*/

ostream &deCpldRotor::print(ostream &os) {        //------------ print ---
  os << "      deCpldRotor " << nSym;
  os.precision(6); os << " " << rotB;
  os.precision(4); os << " " << V0;
  os << " " << sprot << " " << cmntr << endl;
  return os;
}

ostream &operator<<(ostream &os, deCpldRotor &dr)
  { return dr.print(os); }

/*------------------------------------------------------------------------
     Coupled Rotors
------------------------------------------------------------------------*/

ostream &coupledRotors::print(ostream &os) {      //------------ print ---
  int idr, ndr = size(), icr;
  os.setf(ios::fixed);
  os << "    coupledRotors{ " << ndr << endl;
  os.precision(8);
  os << "      qCorrFact " << qCorrFact << endl;
  os << "      idcsRemVib";
  for (icr = 0; icr < ndr; icr++) { os << " " << idcsRemVib[icr]; }
  os << endl;
  os.precision(4);
  os << "      freqRemVib";
  for (icr = 0; icr < ndr; icr++) { os << " " << freqRemVib[icr]; }
  os << endl;
  for (idr = 0; idr < ndr; idr++) { os << (*this)[idr]; }
  os << "    }coupledRotors\n";
  return os;
}

ostream &operator<<(ostream &os, coupledRotors &crs)
  { return crs.print(os); }

/*------------------------------------------------------------------------
     Property modifications for thermo & TST calculations
------------------------------------------------------------------------*/

modProps::modProps() { clear(); }

void modProps::clear() {                               //------- clear ---
  eStates.clear(); isomStates.clear(); modRots.clear(); cpldRots.clear();
}

                                                       //-------- init ---
void modProps::init(jobInfos &ji, molecGeom &mg, freqTable &ft) {
  int nimgFrq = ft.countImag();
  if (nimgFrq > 0) { isTS = true; } else { isTS = false; }
  gElec = mg.gElec; rotSymNbr = mg.rotSymNbr; numIsomers = mg.numIsomers;
  energySCFplsZPE = atof(ji.SCFenergy.c_str()) + ft.zpe / WAVNHRT;
  energy0K = energySCFplsZPE;
}

int modProps::parseInput(string &slin) {          //------- parseInput ---
  vector<string> tkns = splitTokens(slin, " "), stk;
  int ntkns = tkns.size(), old, nes, ies, itk;
  string key = tkns[0];
  eigenState egst;
  listStates *lsts;
  modIntRotor mir;

  if (key == "useSCFenergy+ZPE") { energy0K = energySCFplsZPE;
  } else if (ntkns < 2) { return ERRE;
  } else if (key == "gElec") { gElec = atoi(tkns[1].c_str());
  } else if (key == "rotSymNbr") { rotSymNbr = atof(tkns[1].c_str());
  } else if (key == "numIsomers") { numIsomers = atof(tkns[1].c_str());
  } else if (key == "isTS") { isTS = getBlnvStr(tkns[1]);
  } else if ((key == "energy0K") || (key == "energyTST")) {
    energy0K = atof(tkns[1].c_str());
  } else if (ntkns < 4) { return ERRE;
  } else if ((key == "rotorSpec") || (key == "setIntRotor")) {
    if (key == "setIntRotor") { old = true; } else { old = false; }
    if (old) { if (ntkns < 5) { return ERRE; } }
    mir.idxVib = atoi(tkns[1].c_str()); mir.nSym = atoi(tkns[2].c_str());
    mir.cmntr.erase();
    if (old) {
      stk = splitTokens(tkns[4], "-"); mir.sprot = stk[0] + ":" + tkns[3];
      if (ntkns > 5) { mir.cmntr = tkns[5]; }
    } else {
      mir.sprot = tkns[3]; if (ntkns > 4) { mir.cmntr = tkns[4]; }
    }
    modRots.push_back(mir);
  } else if ((key == "eStates") || (key == "isomStates")) {
    lsts = &eStates; if (key == "isomStates") { lsts = &isomStates; }
    lsts->clear();
    nes = atoi(tkns[1].c_str());
    for (ies = 0; ies < nes; ies++) {
      itk = ies * 2 + 2;
      egst.deg = atoi(tkns[itk].c_str());
      egst.energy = atof(tkns[itk + 1].c_str());
      lsts->push_back(egst);
    }
  } else { cout << "Warning: Invalid key [" << key << "].\n"; }
  return NORM;
}

int modProps::numRotors() {                       //-------- numRotors ---
  int nrots = 0, icr, ncr = cpldRots.size();
  for (icr = 0; icr < ncr; icr++) { nrots += cpldRots[icr].size(); }
  return nrots;
}

int modProps::isIdxRemVib(int idv) {              //------ isIdxRemVib ---
  int icr, ncr = cpldRots.size(), ir, nr;
  for (icr = 0; icr < ncr; icr++) {
    nr = cpldRots[icr].size();
    for (ir = 0; ir < nr; ir++)
      { if (cpldRots[icr].idcsRemVib[ir] == idv) { return true; } }
  }
  return false;
}

ostream &modProps::print(ostream &os) {           //------------ print ---
  int prec, nes, nis, ncr, icr;
  ios::fmtflags flgs;
  prec = os.precision(); flgs = os.flags();
  os.unsetf(ios::fixed); os.precision(6);
  nes = eStates.size(); nis = isomStates.size(); ncr = cpldRots.size();
  os << "modProps{\n";
  os << "  isTS " << strGpoBlnv(isTS) << endl;
  os << "  gElec " << gElec << endl;
  os << "  rotSymNbr " << rotSymNbr << endl;
  os << "  numIsomers " << numIsomers << endl;
  os.setf(ios::fixed); os.precision(8);
  os << "  energy0K " << energy0K << endl;
  os << "  eStates{ " << nes << endl << eStates << "  }eStates\n";
  os << "  isomStates{ " << nis << endl
   << isomStates << "  }isomStates\n";
  os << "  vectorCoupledRotors{ " << ncr << endl;
  for (icr = 0; icr < ncr; icr++) { os << cpldRots[icr]; }
  os << "  }vectorCoupledRotors\n";
  os << "}modProps\n";
  os.flags(flgs); os.precision(prec);
  return os;
}

ostream &operator<<(ostream &os, modProps &mp)
  { return mp.print(os); }

/*========================================================================
     Classes for intramolecular rotation calculation
========================================================================*/

/*------------------------------------------------------------------------
     Top attached on the molecular frame
------------------------------------------------------------------------*/

int attachedTop::setFrom(string dtop, geomInfo &gi) {  //----- setFrom ---
  vector<string> vdtstr = splitTokens(dtop, "|");
  vector<string> tkns = splitTokens(vdtstr[0], ":-"), vatms, vcrds;
  vector<int> vida;
  coordCart crda1, crda2;
  char typc;
  string tkn0;
  int itk, ntk = tkns.size(), namol = gi.size(), idx, ic, ia;

  clear(); idxAtms.clear(); usedIdx.clear();
  if ((ntk < 3) || (ntk > (namol - 1))) { return ERRE; }
  strAllToLower(tkns[0]); typc = tkns[0][0];
  if ((typc == 'x') || (typc == 'y') || (typc == 'z')) {
    axType = typc; tkn0 = tkns[0].substr(1);
  } else if ((typc == 'v') || (typc == 'p')) {
    axType = typc; tkn0 = tkns[0].substr(1);
    if (vdtstr.size() < 2) { return ERRE; }
    vatms = splitTokens(vdtstr[1], ":-");
    if (vatms.size() < 2) { return ERRE; }
    vida.clear();
    for (ia = 0; ia < vatms.size(); ia++) {
      idx = atoi(vatms[ia].c_str()) - 1;
      if ((idx < 0) || (idx >= namol)) { return ERRE; }
      vida.push_back(idx);
    }
    crda1 = gi[vida[1]] - gi[vida[0]];
    if (typc == 'v') { axVctSp = crda1; }
    else {
      if (vatms.size() < 4) { return ERRE; }
      crda2 = gi[vida[3]] - gi[vida[2]];
      axVctSp = vectProd(crda1, crda2);
    }
    axVctSp.normalize();
  } else if (typc == 'i') {
    axType = typc; tkn0 = tkns[0].substr(1);
    if (vdtstr.size() < 2) { return ERRE; }
    vcrds = splitTokens(vdtstr[1], ":");
    if (vcrds.size() < 3) { return ERRE; }
    for (ic = 0; ic < 3; ic++) { axVctSp[ic] = atof(vcrds[ic].c_str()); }
    axVctSp.normalize();
  } else {
    axType = 'z'; tkn0 = tkns[0];
  }
  idxCntPiv = atoi(tkn0.c_str()) - 1;
  if ((idxCntPiv < 0) || (idxCntPiv >= namol)) { return ERRE; }
  usedIdx.insert(idxCntPiv);
  cntPiv = gi[idxCntPiv];
  for (itk = 1; itk < ntk; itk++) {
    idx = atoi(tkns[itk].c_str()) - 1;
    if ((idx < 0) || (idx >= namol)) { return ERRE; }
    if (usedIdx.find(idx) != usedIdx.end()) { return ERRE; }
    usedIdx.insert(idx);
    idxAtms.push_back(idx);
    push_back(gi[idx]);
  }
  molW = gi.molW; molCOM = gi.com; molPrcAxes = gi.prcAxes;
  for (ic = 0; ic < 3; ic++) { molI[ic] = gi.momInert[ic]; }
  return NORM;
}

/* ###
                                                 //--- setFrom(s,s,gi) ---
int attachedTop::setFrom(string dr1, string dr2, geomInfo &gi) {
  // for older file compatibility
  vector<string> tkns = splitTokens(dr2, "-");
  int ntk = tkns.size();
  string dtop;
  if (ntk < 2) { return ERRE; }
  dtop = tkns[0] + ':' + dr1;
  return setFrom(dtop, gi);
}
### */

                                                  //--- beCounterTopOf ---
void attachedTop::beCounterTopOf(attachedTop &ctop, geomInfo &gi) {
  int namol = gi.size(), nactop = ctop.size(), idx, ic;

  clear(); idxAtms.clear(); usedIdx.clear();
  rotorIdx = -ctop.rotorIdx;
  axType = ctop.axType;
  idxCntPiv = ctop.idxAtms[0]; usedIdx.insert(idxCntPiv);
  cntPiv = ctop[0];
  idx = ctop.idxCntPiv; usedIdx.insert(idx); idxAtms.push_back(idx);
  push_back(ctop.cntPiv);
  for (idx = 0; idx < namol; idx++) {
    if (ctop.usedIdx.find(idx) == ctop.usedIdx.end()) {
      usedIdx.insert(idx); idxAtms.push_back(idx);
      push_back(gi[idx]);
    }
  }
  molW = gi.molW; molCOM = gi.com; molPrcAxes = gi.prcAxes;
  for (ic = 0; ic < 3; ic++) { molI[ic] = gi.momInert[ic]; }
}

void attachedTop::setBasicProp() {                //----- setBasicProp ---
  int ia, na = size(), ic, ic_pls, ic_mns, ip;
  double m, lambda;
  coordCart topcrd;

  setTopCOM();
  defineTopAxes();
  // --- re-define top axes when axType != z ---
  if (axType == 'y') {
    cntPiv = (*this)[0] - topAxes.Y(); defineTopAxes();
  } else if (axType == 'x') {
    cntPiv = (*this)[0] - topAxes.X(); defineTopAxes();
  } else if ((axType == 'v') || (axType == 'p') || (axType == 'i')) {
    cntPiv = (*this)[0] - axVctSp; defineTopAxes();
  }

  // --- set r and alpha's with molecular principal axes
  r = molPrcAxes * (topOrigin - molCOM);
  for (ic = 0; ic < 3; ic++) {
    for (ip = 0; ip < 3; ip++)
      { alpha[ic][ip] = topAxes[ic] * molPrcAxes[ip]; }
  }

  // --- calculate A, B, C, and Ux (as in Pitzer-II)
  A = 0.; B = 0.; C = 0.; Ux = 0.;
  for (ia = 1; ia < na; ia++) {
    m = elm.getAtmcWeight((*this)[ia].atomicNumber);
    topcrd = topAxes * ((*this)[ia] - topOrigin);
    A += m * (topcrd.x() * topcrd.x() + topcrd.y() * topcrd.y());
    B += m * topcrd.x() * topcrd.z();
    C += m * topcrd.y() * topcrd.z();
    Ux += m * topcrd.x();
  }

  // --- calculate beta with molecular principal axes
  for (ic = 0; ic < 3; ic++) {
    ic_pls = cycIncCrdIdx(ic); ic_mns = cycDecCrdIdx(ic);
    beta[ic] = alpha.Z(ic) * A - alpha.X(ic) * B - alpha.Y(ic) * C
     + Ux * (alpha.Y(ic_mns) * r[ic_pls] - alpha.Y(ic_pls) * r[ic_mns]);
  }

  // --- I0 is reduced moment of inertia without coupling
  lambda = Ux * Ux / molW;
  for (ic = 0; ic < 3; ic++) {
    if (fabs(molI[ic]) > 1e-10)
     { lambda += (beta[ic] * beta[ic] / molI[ic]); }
  }
  I0 = A - lambda;
  rotC0 = BIWAA2 / I0;
}

double attachedTop::crossLambda(attachedTop &at) {   //--- crossLambda ---
  int ic;
  double clambda = 0.;
  for (ic = 0; ic < 3; ic++) {
    clambda += (alpha.Y(ic) * at.alpha.Y(ic) * Ux * at.Ux / molW);
    if (fabs(molI[ic]) > 1e-10) {
      clambda += (beta[ic] * at.beta[ic] / molI[ic]);
    }
  }
  return clambda;
}

double attachedTop::crossBeta(attachedTop &at) {     //----- crossBeta ---
  int ic, icc, ic_pls, ic_mns;
  double cbeta;
  matrxCart calpha;
  coordCart cr;
  for (ic = 0; ic < 3; ic++) {
    for (icc = 0; icc < 3; icc++)
      { calpha[ic][icc] = topAxes[ic] * at.topAxes[icc]; }
  }
  cr = at.topAxes * (topOrigin - at.topOrigin);
  ic = IZ;
  ic_pls = cycIncCrdIdx(ic); ic_mns = cycDecCrdIdx(ic);
  cbeta = calpha.Z(ic) * A - calpha.X(ic) * B - calpha.Y(ic) * C
   + Ux * (calpha.Y(ic_mns) * cr[ic_pls] - calpha.Y(ic_pls) * cr[ic_mns]);
  return cbeta;
}

int attachedTop::isChildOf(attachedTop &at) {     //------- isChiledOf ---
  set<int>::iterator p;
  if (size() >= at.size()) { return false; }
  for (p = usedIdx.begin(); p != usedIdx.end(); p++) {
    if (at.usedIdx.find(*p) == at.usedIdx.end()) { return false; }
  }
  return true;
}

                                                  //----- setRotDevSim ---
void attachedTop::setRotDevSim(geomInfo &gi, freqTable &ft) {
  rotDevn = getRotDev(gi).normalize();
  simVib = ft.findSimil(rotDevn);
}

devnInfo attachedTop::getRotDev(geomInfo &gi) {   //-------- getRotDev ---
  int ia, natop = size(), ida, namol = gi.size(), iax;
  double r;
  coordCart dev0, vpnt, offAx, dev;
  axisCart axRot;
  devnInfo rotdev;

  dev0.fillZero();
  for (ida = 0; ida < namol; ida++) { rotdev.push_back(dev0); }
  axRot.setByPoints(cntPiv, (*this)[0]);
  for (ia = 1; ia < natop; ia++) {
    vpnt = axRot.vertPoint((*this)[ia]);
    offAx = (*this)[ia] - vpnt;
    if ((r = offAx.vctLength()) >= GAUCPRC) {
      dev = vectProd(axRot.direct, offAx); dev.normalize(); dev *= r;
      rotdev[idxAtms[ia]] = dev;
    }
  }
  rotdev -= gi.getVelCOM(rotdev);
  for (iax = 0; iax < 3; iax++)
    { rotdev -= gi.getAngVPrcAx(rotdev, iax) * gi.getUnitRotPrcAx(iax); }
  return rotdev;
}

                                                  //----- chkHrdwrdIdx ---
void attachedTop::chkHrdwrdIdx(int vidx, freqTable &ft) {
  int ism, nsm = simVib.size();
  devnInfo ref, tgt;
  vibSimil vsm;
  simHrdWrd.clear();
  if (simVib[0].idx == vidx) { return; }
  for (ism = 1; ism < nsm; ism++) {
    if (simVib[ism].idx == vidx)
      { simHrdWrd.push_back(simVib[ism]); return; }
  }
  ref = rotDevn; ref.normalize(); tgt = ft[vidx].nrm; tgt.normalize();
  vsm.idx = vidx; vsm.simil = -fabs(tgt * ref);
  vsm.freq = ft[vidx].freq; simHrdWrd.push_back(vsm);
}

void attachedTop::setTopCOM() {                   //-------- setTopCOM ---
  // --- calc. top center-of-mass excluding pivot
  int ia, na = size(), ic;
  double summ = 0., m;
  topCOM.fillZero();
  for (ia = 1; ia < na; ia++) {
    m = elm.getAtmcWeight((*this)[ia].atomicNumber); summ += m;
    for (ic = 0; ic < 3; ic++) { topCOM[ic] += m * (*this)[ia][ic]; }
  }
  for (ic = 0; ic < 3; ic++) { topCOM[ic] /= summ; }
}

void attachedTop::defineTopAxes() {               //---- defineTopAxes ---
  // --- define asymmetric top axes (as in Pitzer-II)
  coordCart wp;
  axisCart topRotAxis;
  topRotAxis.setByPoints(cntPiv, (*this)[0]);
  topAxes.Z() = topRotAxis.direct;
  topOrigin = topRotAxis.vertPoint(topCOM);
  topAxes.X() = topCOM - topOrigin;
  if (topAxes.X().vctLength() < GAUCPRC) {
    wp = (*this)[1]; topAxes.X() = wp - topRotAxis.vertPoint(wp);
    if (topAxes.X().vctLength() < GAUCPRC) {
      wp = topOrigin; wp[topAxes.Z().minAbsIndex()] += 1.;
      topAxes.X() = wp - topRotAxis.vertPoint(wp);
    }
  }
  topAxes.X().normalize();
  topAxes.Y() = vectProd(topAxes.Z(), topAxes.X());
}

/*------------------------------------------------------------------------
     Molecule with attached tops
------------------------------------------------------------------------*/

void molecWithTops::clear() {                     //------------ clear ---
  aTops.clear(); molec.clear(); moI.clear(); roB.clear(); coefCpl.clear();
  dvnCoupled.clear(); simCoupled.clear(); freqDeCpl.clear();
  worseCpld.clear(); dupVbAsgn.clear(); imgAsgnd.clear();
}

void molecWithTops::setGeom(geomInfo &gi)         //---------- setGeom ---
  { molec = gi; molec.setPrcAxesMomInert(); }

                                                  //----- setOneTopIdx ---
int molecWithTops::setOneTopIdx(string s, freqTable &ft, int ridx) {
  attachedTop top;
  if (top.setFrom(s, molec) == ERRE) { return ERRE; }
  top.rotorIdx = ridx; top.setBasicProp(); top.setRotDevSim(molec, ft);
  aTops.push_back(top);
  return NORM;
}

                                    // --- setTops (without freqTable) ---
int molecWithTops::setTops(vector<string> &vs) {
  int itp, ntp = vs.size();
  attachedTop top;
  aTops.clear();
  for (itp = 0; itp < ntp; itp++) {
    if (top.setFrom(vs[itp], molec) == ERRE) { return ERRE; }
    top.rotorIdx = itp; top.setBasicProp();
    aTops.push_back(top);
  }
  groupIdx = 0;
  return NORM;
}
                                       // --- setTops (with freqTable) ---
int molecWithTops::setTops(vector<string> &vs, freqTable &ft) {
  int itp, ntp = vs.size();
  attachedTop top;
  aTops.clear();
  for (itp = 0; itp < ntp; itp++) {
    if (top.setFrom(vs[itp], molec) == ERRE) { return ERRE; }
    top.rotorIdx = itp; top.setBasicProp(); top.setRotDevSim(molec, ft);
    aTops.push_back(top);
  }
  groupIdx = 0;
  return NORM;
}

                               // --- calcCoupling (without freqTable) ---
void molecWithTops::calcCoupling() {
  const int lw = 1000;
  int irow, icol, irt, nrt = aTops.size(), info;
  double mat[MAXSMAT][MAXSMAT], w[MAXSMAT], work[lw];
  double piIuc = 1., piI = 1.;

  for (irow = 0; irow < nrt; irow++) {
    mat[irow][irow] = aTops[irow].I0;
    for (icol = irow + 1; icol < nrt; icol++) {
      mat[icol][irow] = -aTops[irow].crossLambda(aTops[icol]);
      if (aTops[icol].isChildOf(aTops[irow]))
        { mat[icol][irow] += aTops[icol].crossBeta(aTops[irow]); }
      else if (aTops[irow].isChildOf(aTops[icol]))
        { mat[icol][irow] += aTops[irow].crossBeta(aTops[icol]); }
    }
  }
  info = lapack_dsyev('V', 'U', nrt, (double **)mat, MAXSMAT, w,
                      work, lw);
  cout << " DSYEV info:" << info << "  " << "lw-opt:" << work[0] << endl;
  moI.init(nrt); roB.init(nrt); coefCpl.init(nrt, nrt);
  for (icol = 0; icol < nrt; icol++) {
    moI[icol] = w[icol]; roB[icol] = BIWAA2 / moI[icol];
    for (irow = 0; irow < nrt; irow++)
      { coefCpl[icol][irow] = mat[icol][irow]; }
  }
  
  for (irt = 0; irt < nrt; irt++)
    { piIuc *= aTops[irt].I0; piI *= moI[irt]; }
  qCorrFact = sqrt(piI / piIuc);
}

                                  // --- calcCoupling (with freqTable) ---
void molecWithTops::calcCoupling(freqTable &ft) {
  const int lw = 1000;
  int irow, icol, irt, nrt = aTops.size(), info;
  double mat[MAXSMAT][MAXSMAT], w[MAXSMAT], work[lw];
  double piIuc = 1., piI = 1.;

  for (irow = 0; irow < nrt; irow++) {
    mat[irow][irow] = aTops[irow].I0;
    for (icol = irow + 1; icol < nrt; icol++) {
      mat[icol][irow] = -aTops[irow].crossLambda(aTops[icol]);
      if (aTops[icol].isChildOf(aTops[irow]))
        { mat[icol][irow] += aTops[icol].crossBeta(aTops[irow]); }
      else if (aTops[irow].isChildOf(aTops[icol]))
        { mat[icol][irow] += aTops[irow].crossBeta(aTops[icol]); }
    }
  }
  info = lapack_dsyev('V', 'U', nrt, (double **)mat, MAXSMAT, w,
                      work, lw);
  cout << " DSYEV info:" << info << "  " << "lw-opt:" << work[0] << endl;
  moI.init(nrt); roB.init(nrt); coefCpl.init(nrt, nrt);
  for (icol = 0; icol < nrt; icol++) {
    moI[icol] = w[icol]; roB[icol] = BIWAA2 / moI[icol];
    for (irow = 0; irow < nrt; irow++)
      { coefCpl[icol][irow] = mat[icol][irow]; }
  }
  
  for (irt = 0; irt < nrt; irt++)
    { piIuc *= aTops[irt].I0; piI *= moI[irt]; }
  qCorrFact = sqrt(piI / piIuc);
  setDevSimCoupled(ft);
  diagnCoupling();
}

void molecWithTops::calcFreqDeCpl() {             //---- calcFreqDeCpl ---
  int icr, irt, nrt = aTops.size();
  double frq;
  gnVector kf(nrt), kfdeCpl;
  gnMatrix deCpl = (~coefCpl).sqElement();
  for (icr = 0; icr < nrt; icr++) {
    frq = simCoupled[icr][0].freq; kf[icr] = frq * frq / (2. * roB[icr]);
  }
  kfdeCpl = deCpl * kf;
  freqDeCpl.init(nrt);
  for (irt = 0; irt < nrt; irt++)
    { freqDeCpl[irt] = sqrt(2. * aTops[irt].rotC0 * kfdeCpl[irt]); }
}

int molecWithTops::diagnOutput(set<int> &asv) {   //------ diagnOutput ---
  int irg, icr, nrg = aTops.size(), gidx, ridx, nshw, vidx;
  double frq;
  set<int>::iterator p;

  gidx = groupIdx;
  if (nrg < 2) {                             // ----- single rotor --
    ridx = aTops[0].rotorIdx;
    cout << " g:" << gidx << " r:" << ridx + 1;
    if (gidx > 0) {                // --- hardwired --
      vidx = gidx - 1;
      cout << " hard:" << vidx + 1;
      nshw = aTops[0].simHrdWrd.size();
      if (nshw > 0) {
        cout << " sim:" << aTops[0].simHrdWrd[0].simil
         << " ?:" << aTops[0].simVib[0].idx + 1
         << " sim:" << aTops[0].simVib[0].simil << endl;
        frq = aTops[0].simHrdWrd[0].freq;
      } else {
        cout << " sim:" << aTops[0].simVib[0].simil << endl;
        frq = aTops[0].simVib[0].freq;
      }
    } else {                       // -------- auto --
      vidx = aTops[0].simVib[0].idx; frq = aTops[0].simVib[0].freq;
      cout << " auto:" << vidx + 1;
      cout << " sim:" << aTops[0].simVib[0].simil << endl;
    }
    if ((p = asv.find(vidx)) != asv.end())
      { cout << "Duplicated vibIdx.\n"; return ERRE; }
    else { asv.insert(vidx); }
    if (frq < 0.) {
      cout << "Warning: assigned imag. freq: vidx=" << vidx + 1 << endl;
    }
  } else {                                   // ---- coupled rotor --
    cout << " --- group " << gidx << " uncoupled rotors ---\n";
    for (irg = 0; irg < nrg; irg++) {
      ridx = aTops[irg].rotorIdx;
      cout << " g:" << gidx << " r:" << ridx + 1;
      if (worseCpld[irg]) {
        cout << " ?:" << aTops[irg].simVib[0].idx + 1 << " sim:"
         << aTops[irg].simVib[0].simil << " *";
      }
      cout << endl;
    }
    if (someWorseCpld)
      { cout << "Warning: *sim. is worse when coupled.\n"; }
    cout << " --- group " << gidx << " coupling result ---\n";
    for (icr = 0; icr < nrg; icr++) {
      vidx = simCoupled[icr][0].idx; frq = simCoupled[icr][0].freq;
      cout << " cpld:" << icr + 1 << " auto:"
       << simCoupled[icr][0].idx + 1 << " sim:"
       << simCoupled[icr][0].simil << " vec:";
      for (irg = 0; irg < nrg; irg++)
        { cout << " " << coefCpl[icr][irg]; }
      cout << endl;
      if ((p = asv.find(vidx)) != asv.end())
        { cout << "Duplicated vibIdx.\n"; return ERRE; }
      else { asv.insert(vidx); }
      if (imgAsgnd[icr]) {
        cout << "Warning: assigned imag. freq: vidx=" << vidx + 1 << endl;
      }
    }
  }
  return NORM;
}

                                               //---- setDevSimCoupled ---
void molecWithTops::setDevSimCoupled(freqTable &ft) {
  int irt, nrt = aTops.size(), icr;
  coordCart zcrd;
  devnInfo di;
  vector<devnInfo> uncpld;
  uncpld.clear(); dvnCoupled.clear(); simCoupled.clear();
  for (irt = 0; irt < nrt; irt++)
    { uncpld.push_back(aTops[irt].getRotDev(molec)); }
  for (icr = 0; icr < nrt; icr++) {
    di.clear(); di.resize(molec.size(), zcrd);
    for (irt = 0; irt < nrt; irt++)
      { di += uncpld[irt] * coefCpl[icr][irt]; }
    di.normalize(); dvnCoupled.push_back(di);
    simCoupled.push_back(ft.findSimil(di));
  }
}

void molecWithTops::diagnCoupling() {             //---- diagnCoupling ---
  /* diagnose the coupling result */
  int irt, nrt = aTops.size(), icr, idv;
  map<int, int> asgndVib;
  map<int, int>::iterator p;
  someWorseCpld = false;
  worseCpld.clear(); worseCpld.resize(nrt, false);
  for (irt = 0; irt < nrt; irt++) {
    if (simCplLrgFor(irt) < aTops[irt].simVib[0].simil)
      { worseCpld[irt] = true; someWorseCpld = true; }
  }
  someDupVbAsgn = false; someImgAsgnd = false;
  dupVbAsgn.clear(); dupVbAsgn.resize(nrt, false);
  imgAsgnd.clear(); imgAsgnd.resize(nrt, false);
  for (icr = 0; icr < nrt; icr++) {
    idv = simCoupled[icr][0].idx;
    if ((p = asgndVib.find(idv)) != asgndVib.end()) {
      dupVbAsgn[icr] = true; dupVbAsgn[p->second] = true;
      someDupVbAsgn = true;
    } else { asgndVib.insert(make_pair(idv, icr)); }
    if (simCoupled[icr][0].freq < 0.)
      { imgAsgnd[icr] = true; someImgAsgnd = true; }
  }
}

double molecWithTops::simCplLrgFor(int irt) {     //----- simCplLrgFor ---
  /* find the coupled rotor with largest coefficient
    for #irt and returns the top similarity with vibrations */
  int icr, nrt = aTops.size();
  double maxCoef = 0., sim = 0.;
  for (icr = 0; icr < nrt; icr++) {
    if (fabs(coefCpl[icr][irt]) > maxCoef) {
      maxCoef = fabs(coefCpl[icr][irt]);
      sim = simCoupled[icr][0].simil;
    }
  }
  return sim;
}

/*------------------------------------------------------------------------
     Coupled internal rotor group
------------------------------------------------------------------------*/

                                                  //---------- setFrom ---
int groupedTops::setFrom(modProps &mp, geomInfo &gi, freqTable &ft) {
  int irm, nrm = mp.modRots.size(), igp, gidx, ngp;
  molecWithTops moltop;
  clear();
  for (irm = 0; irm < nrm; irm++) {
    gidx = mp.modRots[irm].idxVib; igp = findGroupIdx(gidx);
    if (igp < 0) {
      moltop.clear(); moltop.setGeom(gi); moltop.groupIdx = gidx;
      push_back(moltop); igp = size() - 1;
    }
    if ((*this)[igp].setOneTopIdx(mp.modRots[irm].sprot, ft, irm) == ERRE)
      { return ERRE; }
  }
  ngp = size();
  for (igp = 0; igp < ngp; igp++) {
    if ((gidx = (*this)[igp].groupIdx) > 0) {
      if ((*this)[igp].aTops.size() > 1)
        { cout << "Duplicated hardwired vibIdx.\n"; return ERRE; }
      (*this)[igp].aTops[0].chkHrdwrdIdx(gidx - 1, ft);
    }
    (*this)[igp].calcCoupling(ft); (*this)[igp].calcFreqDeCpl();
  }
  return NORM;
}

int groupedTops::diagnOutput() {                  //------ diagnOutput ---
  int igp, ngp = size();
  set<int> asgndVib;
  for (igp = 0; igp < ngp; igp++)
    { if ((*this)[igp].diagnOutput(asgndVib) == ERRE) { return ERRE; } }
  return NORM;
}

int groupedTops::setBackTo(modProps &mp) {        //-------- setBackTo ---
  int igp, ngp = size(), irg, nrg, ridx, icr, vidx;
  double frq;
  coupledRotors crts;
  deCpldRotor dcr;
  mp.cpldRots.clear();
  for (igp = 0; igp < ngp; igp++) {
    nrg = (*this)[igp].aTops.size();
    crts.clear(); crts.idcsRemVib.clear(); crts.freqRemVib.clear();
    for (irg = 0; irg < nrg; irg++) {
      ridx = (*this)[igp].aTops[irg].rotorIdx;
      dcr.nSym = mp.modRots[ridx].nSym;
      dcr.sprot = mp.modRots[ridx].sprot;
      dcr.cmntr = mp.modRots[ridx].cmntr;
      dcr.rotB = (*this)[igp].aTops[irg].rotC0;
      frq = (*this)[igp].freqDeCpl[irg];
      dcr.V0 = (frq / dcr.nSym) * (frq / dcr.nSym) / dcr.rotB;
      crts.push_back(dcr);
    }
    crts.qCorrFact = (*this)[igp].qCorrFact;
    for (icr = 0; icr < nrg; icr++) {
      vidx = (*this)[igp].simCoupled[icr][0].idx;
      if ((mp.isTS) && (vidx == 0))
        { cout << "An intRot assigned to TS mode.\n"; return ERRE; }
      crts.idcsRemVib.push_back(vidx);
      crts.freqRemVib.push_back((*this)[igp].simCoupled[icr][0].freq);
    }
    mp.cpldRots.push_back(crts);
  }
  return NORM;
}

int groupedTops::findGroupIdx(int gidx) {         //----- findGroupIdx ---
  int igp, ngp = size();
  for (igp = 0; igp < ngp; igp++)
    { if ((*this)[igp].groupIdx == gidx) { return igp; } }
  return -1;
}

/*========================================================================
     global library functions
========================================================================*/

/*----- output .gpo file -----------------------------------------------*/

int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modProps &mp) {
  string gpofln = basfln + ".gpo";
  ofstream gpof;
  gpof.open(gpofln.c_str());
  if (!gpof)
    { cout << "failed to open [" << gpofln << "].\n"; return ERRE; }
  gpof << "# gpop link data file created from " << basfln << ".mod.\n";
  gpof << ji << mg << ft << mp;
  gpof.close();
  return NORM;
}

/*----- output .tst file -----------------------------------------------*/

int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modProps &mp) {
  int prec, nes, nis, irot, nfrq, ifrq, ncr, icr, nr, ir;
  ios::fmtflags flgs;
  string toutfln = basfln + ".tst";
  ofstream tof;

  tof.open(toutfln.c_str());
  if (!tof)
    { cout << "failed to open [" << toutfln << "].\n"; return ERRE; }
  prec = tof.precision(); flgs = tof.flags(); tof.setf(ios::fixed);
  tof << "[Name of the Molecule]\n";
  tof << " " << ji.comment << endl;
  tof << "[Molecular Mass(amu)]\n";
  tof.precision(6); tof << " " << mg.molW << endl;
  tof << "[Deg Electron State |neg|=NumState followed by (gi, ei(cm-1))'s]\n";
  if ((nes = mp.eStates.size()) > 0)
    { tof << " " << (-nes) << endl << mp.eStates.outStrTST(); }
  else { tof << " " << mp.gElec << endl; }
  tof << "[Rot Sym Num / Num of Isomrs]\n";
  tof << " -1\n " << (int)mp.rotSymNbr << " ";
  if ((nis = mp.isomStates.size()) > 0)
    { tof << nis << endl << mp.isomStates.outStrTST(); }
  else { tof << "1\n " << (int)mp.numIsomers << " 0.\n"; }
  tof << "[Num of RotConsts (1:linear 3:nonlinear), RotConsts(GHz)]\n";
  tof.precision(7);
  if (mg.nAtms == 1) {
    tof << " 0\n";
  } else {
    if (mg.isLinear) {
      tof << " 1\n";
      tof << " " << mg.rotcGHz[1] << endl;
    } else {
      tof << " 3\n";
      for (irot = 0; irot < 3; irot++)
        { tof << " " << mg.rotcGHz[irot] << endl; }
    }
  }
  tof << "[Num of Freqs, Freqs(cm-1) |neg|=1D-rotorB(cm-1)]\n";
  nfrq = mg.nVibm - mp.numRotors(); if (mp.isTS) { nfrq--; }
  tof << " " << nfrq << endl;
  tof.precision(4);
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    if ((!mp.isTS) || ifrq > 0) {
      if (!mp.isIdxRemVib(ifrq))
        { tof << " " << ft[ifrq].freq * ji.scaleFactVib << endl; }
    }
  }
  tof << "[-Num.Cpl.R. / nR frqs [B(cm-1), V0(cm-1), SigN]]\n";
  ncr = mp.cpldRots.size();
  tof << " " << -ncr << endl;
  for (icr = 0; icr < ncr; icr++) {
    nr = mp.cpldRots[icr].size();
    tof.precision(4); tof << "  " << nr;
    for (ir = 0; ir < nr; ir++)
      { tof << " " << mp.cpldRots[icr].freqRemVib[ir]; }
    tof << endl;
    for (ir = 0; ir < nr; ir++) {
      tof.precision(6); tof << "   " << mp.cpldRots[icr][ir].rotB;
      tof.precision(4); tof << " " << mp.cpldRots[icr][ir].V0;
      tof << " " << mp.cpldRots[icr][ir].nSym << endl;
    }
  }
  tof << "[energy 0K for TST (hartree)]\n";
  tof.precision(8); tof << " " << mp.energy0K << endl;
  if (mp.isTS) {
    tof << "[imaginary frequency (cm-1) .. only for TS molecule]\n";
    tof.precision(4); tof << " " << fabs(ft[0].freq) << endl;
  }
  tof.flags(flgs); tof.precision(prec);
  tof.close();
  return NORM;
}

