/*========================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2013 by A. Miyoshi, Univ. Tokyo
========================================================================*/

#ifndef _GPOPLIBR_H_
#define _GPOPLIBR_H_

#include "gpoplibc.h"

/*========================================================================
     Classes for molecular states
========================================================================*/

/*------------------------------------------------------------------------
     Eigen state
------------------------------------------------------------------------*/

class eigenState {
 public:                             // --- data members ---
  int deg;          // degeneracy
  double energy;    // energy [cm-1]
 public:                         // --- member functions ---
  eigenState();
  void clear();
  int restoreFrom(istream &is);
  int restoreFromTST(istream &is);
  ostream &print(ostream &os);
  string outStrTST();
};                                        // --- globals ---
ostream &operator<<(ostream &os, eigenState &es);

/*------------------------------------------------------------------------
     List of eigenstates
------------------------------------------------------------------------*/

class listStates : public vector<eigenState> {
 public:                         // --- member functions ---
  int restoreFrom(istream &is, int nes);
  int restoreFromTST(istream &is, int nes);
  void setOneState(int deg);
  ostream &print(ostream &os);
  string outStrTST();
  double statesQ(double T);
  double statesU(double T);
  double statesS(double T);
  double statesCv(double T);
 private:
  double statesSum(int n, double T);
};                                        // --- globals ---
ostream &operator<<(ostream &os, listStates &ls);


/*========================================================================
     Classes for molecular-property modification
========================================================================*/

/*------------------------------------------------------------------------
     Modification to internal rotor
------------------------------------------------------------------------*/

class modIntRotor {
 public:                             // --- data members ---
  int idxVib, nSym;
  string sprot;       // specification string of rotor atoms
  string cmntr;       // comment on the rotor specification
};

/*------------------------------------------------------------------------
     Decoupled Rotor
------------------------------------------------------------------------*/

class deCpldRotor {
 public:                             // --- data members ---
  int nSym;
  double rotB, V0;    // [cm-1]
  string sprot;       // specification string of rotor atoms
  string cmntr;       // comment on the rotor specification
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, deCpldRotor &dr);

/*------------------------------------------------------------------------
     Coupled Rotors
------------------------------------------------------------------------*/

class coupledRotors : public vector<deCpldRotor> {
 public:                             // --- data members ---
  double qCorrFact;
  vector<int> idcsRemVib;
  vector<double> freqRemVib;
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, coupledRotors &crs);

/*------------------------------------------------------------------------
     Property modifications for thermo & TST calculations
------------------------------------------------------------------------*/

class modProps {
 public:                             // --- data members ---
  int isTS, gElec;
  double rotSymNbr, numIsomers;
  double energy0K;       // [hartree]
  listStates eStates, isomStates;
  vector<modIntRotor> modRots;
  vector<coupledRotors> cpldRots;
 private:
  double energySCFplsZPE;
 public:                         // --- member functions ---
  modProps();
  void clear();
  void init(jobInfos &ji, molecGeom &mg, freqTable &ft);
  int parseInput(string &slin);
  int numRotors();
  int isIdxRemVib(int idv);
//#####  int restoreFrom(istream &is); //#### to be done
  ostream &print(ostream &os);
};                                        // --- globals ---
ostream &operator<<(ostream &os, modProps &mp);


/*========================================================================
     Classes for intramolecular rotation calculation
========================================================================*/

/*------------------------------------------------------------------------
     Top attached on the molecular frame
------------------------------------------------------------------------*/

class attachedTop : public vector<atomCoord> {
 public:                             // --- data members ---
  int rotorIdx;
  char axType;
  coordCart axVctSp;
  atomCoord cntPiv;
  vector<int> idxAtms;
  int idxCntPiv;
  set<int> usedIdx;
  double molW;
  coordCart molCOM;
  double molI[3];
  matrxCart molPrcAxes;
    // --- basic top properties
  coordCart topCOM;         // top center-of-mass excluding pivots
  coordCart topOrigin;      // origin of top axes
  matrxCart topAxes;        // top axes (in original coordinate)
  coordCart r;              // topOrigin - molCOM (in prnc axes coord)
  matrxCart alpha;          // top axes (in principal axes coordinate)
  double A, B, C, Ux, I0, rotC0;
  coordCart beta;
    // --- extended properties
  devnInfo rotDevn;
  vibSimList simVib;
  vibSimList simHrdWrd;
 public:                         // --- member functions ---
  int setFrom(string dtop, geomInfo &gi);
// ###  int setFrom(string dr1, string dr2, geomInfo &gi);
  void beCounterTopOf(attachedTop &ctop, geomInfo &gi);
  void setBasicProp();
  double crossLambda(attachedTop &at);
  double crossBeta(attachedTop &at);
  int isChildOf(attachedTop &at);
  void setRotDevSim(geomInfo &gi, freqTable &ft);
  devnInfo getRotDev(geomInfo &gi);
  void chkHrdwrdIdx(int vidx, freqTable &ft);
 private:
  void setTopCOM();
  void defineTopAxes();
};

/*------------------------------------------------------------------------
     Molecule with attached tops
------------------------------------------------------------------------*/

class molecWithTops {
 public:                             // --- data members ---
  int groupIdx;
  vector<attachedTop> aTops;
  geomInfo molec;
    // --- coupling properties
  gnVector moI, roB;
  gnMatrix coefCpl;
  vector<devnInfo> dvnCoupled;
  vector<vibSimList> simCoupled;
  double qCorrFact;
  gnVector freqDeCpl;
  vector<int> worseCpld, dupVbAsgn, imgAsgnd;
  int someWorseCpld, someDupVbAsgn, someImgAsgnd;
 public:                         // --- member functions ---
  void clear();
  void setGeom(geomInfo &gi);
  int setOneTopIdx(string s, freqTable &ft, int ridx);
  int setTops(vector<string> &vs);
  int setTops(vector<string> &vs, freqTable &ft);
  void calcCoupling();
  void calcCoupling(freqTable &ft);
  void calcFreqDeCpl();
  int diagnOutput(set<int> &asv);
 private:
  void setDevSimCoupled(freqTable &ft);
  void diagnCoupling();
  double simCplLrgFor(int irt);
};

/*------------------------------------------------------------------------
     Coupled internal rotor group
------------------------------------------------------------------------*/

class groupedTops : public vector<molecWithTops> {
 public:                         // --- member functions ---
  int setFrom(modProps &mp, geomInfo &gi, freqTable &ft);
  int diagnOutput();
  int setBackTo(modProps &mp);
 private:
  int findGroupIdx(int gidx);
};

/*========================================================================
     global library functions
========================================================================*/

// int readGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
//  modProps &mp);
int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modProps &mp);
int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modProps &mp);

#endif

