/*========================================================================
     "numlibArf" - regression to Arrhenius expression
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
                                  created: Dec.  29, 2009
                              last edited: Dec.  29, 2009
========================================================================*/

#include "numlibArf.h"

/*========================================================================
     Rate coefficients regression to Arrhenius & modified Arrhenius
      expressions
========================================================================*/

fitArrhenius::fitArrhenius() { clear(); }          // --- initializers ---

void fitArrhenius::clear() {
  basicLinLsSq::clear();
  kdat.clear();
  A = 0.; EaR = 0.; mA = 0.; mb = 0.; mEaR = 0.;
  sdk = 0.; madk = 0.; maxadk = 0.; msdk = 0.; mmadk = 0.; mmaxadk = 0.;
}

                                             // --- add one k(T) datum ---
int fitArrhenius::addDatumKT(double k, double T) {
  kDatumT kd;
  if (k <= 0.) { cout << "invalid k.\n"; return ERRE; }
  if (T <= 0.) { cout << "invalid T.\n"; return ERRE; }
  kd.k = k; kd.T = T; kdat.push_back(kd);
  return NORM;
}

                                                  // --- add k(T) data ---
int fitArrhenius::addDataKT(vector<kDatumT> &vkd) {
  int i, n = vkd.size();
  for (i = 0; i < n; i++)
    { if (addDatumKT(vkd[i].k, vkd[i].T) == ERRE) { return ERRE; } }
  return NORM;
}

int fitArrhenius::solve() {                               // --- solve ---
  int i, nd = kdat.size();
  double y, sdy, adk;
  vector<double> x, mlp, sdp;
  vector<vector<double> > pcc;
  if (nd < 2) { cout << "Too small number of data.\n"; return ERRE; }

  // - Arrhenius fit
  if (init(2) == ERRE) { return ERRE; }
  x.clear(); x.push_back(1.); x.push_back(0.);
  for (i = 0; i < nd; i++) {
    x[1] = 1. / kdat[i].T; y = log(kdat[i].k);
    if (addDatum(y, x) == ERRE) { return ERRE; }
  }
  if (solve_blls(mlp, sdp, pcc, sdy) == ERRE) { return ERRE; }
  A = exp(mlp[0]); EaR = -mlp[1]; sdk = sdy;
  madk = 0.; maxadk = 0.;
  for (i = 0; i < nd; i++) {
    adk = fabs(mlp[0] + mlp[1] / kdat[i].T - log(kdat[i].k));
    madk += adk;
    if (adk > maxadk) { maxadk = adk; }
  }
  madk /= nd;

  // - modified Arrhenius fit
  mA = 0.; mb = 0.; mEaR = 0.; msdk = 0.; mmadk = 0.; mmaxadk = 0.;
  if (nd < 3) {
    cout << "Too small number of data for modified Arrhenius.\n";
    return NORM;
  }
  if (init(3) == ERRE) { return ERRE; }
  x.clear(); x.push_back(1.); x.push_back(0.); x.push_back(0.);
  for (i = 0; i < nd; i++) {
    x[1] = log(kdat[i].T); x[2] = 1. / kdat[i].T; y = log(kdat[i].k);
    if (addDatum(y, x) == ERRE) { return ERRE; }
  }
  if (solve_blls(mlp, sdp, pcc, sdy) == ERRE) { return ERRE; }
  mA = exp(mlp[0]); mb = mlp[1]; mEaR = -mlp[2]; msdk = sdy;
  mmadk = 0.; mmaxadk = 0.;
  for (i = 0; i < nd; i++) {
    adk = fabs(mlp[0] + mlp[1] * log(kdat[i].T) + mlp[2] / kdat[i].T - log(kdat[i].k));
    mmadk += adk;
    if (adk > mmaxadk) { mmaxadk = adk; }
  }
  mmadk /= nd;
  return NORM;
}

